/**
 * Copyright 2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 */
package com.sunlabs.archsim.fleet.ships.bitbucket;

import java.util.LinkedList;
import java.util.List;

import com.sunlabs.archsim.base.common.Parameters;
import com.sunlabs.archsim.base.simulation.Delay;
import com.sunlabs.archsim.fleet.common.Data;
import com.sunlabs.archsim.fleet.common.FleetGlobal;
import com.sunlabs.archsim.fleet.common.Payload;
import com.sunlabs.archsim.gasp.Action;
import com.sunlabs.archsim.gasp.Component;
import com.sunlabs.archsim.gasp.Function;
import com.sunlabs.archsim.gasp.Terminal;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;
import com.sunlabs.archsim.globals.GlobalDefaults;

/**
 * @author ib27688
 *
 */
public class BitBucket extends Component {
	
	protected Terminal in;
	protected Terminal ackIn;
	
	protected BitBucketAction bbAction;
	
	// the list to keep data
	List<Object> receivedData= new LinkedList<Object>();
	
	public BitBucket(String name, Parameters params) {
		super(name, null);

		Delay ackDelay= null;
		boolean keep= BitBucketDefaults.BIT_BUCKET_KEEP_DATA_DEF;
		boolean setKeep= false;
		
		if (params != null) {
			for (int i= 0; i<params.size(); i++) {
				String pName= params.getName(i);
				String sValue= params.getValue(i);
				if (pName.equals(BitBucketDefaults.BIT_BUCKET_DELAY)) {
					try {
						int d= Integer.parseInt(sValue);
						ackDelay= new Delay(d);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for "
								+ BitBucketDefaults.BIT_BUCKET_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED, default value of "
								+ ackDelay 
								+ " is used.");						
					}					
				}
				else if (pName.equals(BitBucketDefaults.BIT_BUCKET_KEEP_DATA)) {
					try {
						keep= Boolean.parseBoolean(sValue);
						setKeep= true;
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for "
								+ BitBucketDefaults.BIT_BUCKET_KEEP_DATA
								+ ": " + sValue
								+ ". Value is IGNORED, default value of "
								+ keep 
								+ " is used.");						
					}					
					
				}
				else {
					director.error("Component: " + name + ", id= " + getID() 
							+ ": Attribute " + pName + " with value " + sValue
							+ " is ignored");
				}
			} // for
		} // if

		constructBitBucket();

		// take parameters into account
		if (setKeep) {
			bbAction.keepData= keep;
		}
		if (ackDelay != null) {
			bbAction.ackFunction.setDelay(ackDelay);
		}
	} // BitBucket constructor
	
	
	protected void constructBitBucket() {
		// make the action
		bbAction= new BitBucketAction(name+"Action");
		// add the action to the component
		addMember(bbAction);
		
		// make the terminals, add them to the component
		in= new InputTerminal(FleetGlobal.IN, this);
		ackIn= new OutputTerminal(FleetGlobal.ACK_IN, this);

		// attach the terminals
		bbAction.attach(in);
		bbAction.attach(ackIn);		
	}
	

	
	public Terminal getTerminal(String name) {
		if (name.equals(in.getBaseName()) 
				|| name.equals(in.getBaseAlias()))
			return in;
		else if (name.equals(ackIn.getBaseName()) 
				|| name.equals(ackIn.getBaseAlias()))
			return ackIn;
		else 
			return null;
	} // getTerminal
	
	
	public String stateInfo() {
		
		
		return "BitBucket " + getName() + ", id= " + getID() + ",  has received " 
		+ bbAction.counter + " items: " + receivedData;
	} // stateInfo
 
	    
    private class BitBucketAction extends Action {

    	public boolean keepData= false;
    	
    	public int counter= 0;
    	
    	public ActionFired ackFunction;
    	
        public BitBucketAction(String name) {
            super(name);
            // should the data be kept
            Boolean kd= globals.booleanValue(BitBucketDefaults.BIT_BUCKET_KEEP_DATA);
			keepData= (kd != null)
				? kd.booleanValue()
				: BitBucketDefaults.BIT_BUCKET_KEEP_DATA_DEF;    
			ackFunction= new ActionFired();
            outputFunctions.put(FleetGlobal.ACK_IN, ackFunction);
            
        } // FifoAction

        /** Vacuous function used to acknowldge that the action has fired */ 
        private class ActionFired extends Function {

           	public ActionFired() {
                super("BitBucketShipAction");
				Integer fDelay= globals.intValue(BitBucketDefaults.BIT_BUCKET_DELAY);
				functionDelay= (fDelay != null)
					? new Delay(fDelay.intValue())
					: BitBucketDefaults.BIT_BUCKET_DELAY_DEF;                    
    		} // ActionFired

            public Object value() {
               // get data
            	counter++;
            	Object rd= in.getData();
           		// should data be stored?
        		if (keepData) {
        			receivedData.add(rd);
        		}
            	// is it a payload? yes - everything OK
        		if (rd instanceof Payload) {
                	if (GlobalDefaults.JOURNAL_ACTION_DEF ) {
                		director.journal("DataSink " 
                				+ BitBucket.this.getName() 
                				+ " received data " + rd);
                    }            			
        		}
        		else {
            		// not a payload, log if necesary and report error
                	if (GlobalDefaults.JOURNAL_ACTION_DEF ) {
                		director.journal("DataSink " 
                				+ BitBucket.this.getName() 
                				+ " received data " + rd
                				+ " non-payload type "
                				+ rd.getClass().getName());
                    }            			

                	director.error("BitBucket " 
            				+ BitBucket.this.getName() + BitBucket.this.getID()  
            				+ " received a non-payload object: " + rd
            				+ " of type " + rd.getClass().getName());
            	}
            	
            	return new Data();
            } // value
            
        } // ActionFired
    }


	
} // BitBucket
