package edu.berkeley.sbp.meta;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.chr.*;
import edu.berkeley.sbp.misc.*;
import edu.berkeley.sbp.bind.*;
import java.util.*;
import java.lang.annotation.*;
import java.lang.reflect.*;
import java.io.*;

public class MetaGrammarBindings {
    public static @bind class Grammar {
        public NonTerminal get(String s) {
            for(NonTerminal nt : nonterminals)
                if (nt.name.equals(s)) return nt;
            return null;
        }
        public @bind.arg NonTerminal[] nonterminals;
        public String toString() {
            String ret = "[ ";
            for(NonTerminal nt : nonterminals) ret += nt + ", ";
            return ret + " ]";
        }
    }
    public abstract static class Un extends El {
        public Seq[][] sequences;
        public void build(MetaGrammar.Context cx, Union u) {
            HashSet<Sequence> bad2 = new HashSet<Sequence>();
            for(int i=0; i<sequences.length; i++) {
                Seq[] group = sequences[i];
                Union u2 = new Union();
                if (sequences.length==1) u2 = u;
                for(int j=0; j<group.length; j++) {
                    group[j].build(cx, u2, false);
                }
                if (sequences.length==1) break;
                Sequence seq = Sequence.singleton(u2);
                for(Sequence s : bad2) {
                    s.lame = true;
                    seq = seq.not(s);
                }
                u.add(seq);
                bad2.add(Sequence.singleton(u2));
            }
        }
    }
    public static class NonTerminal extends Un {
        public String  name = null;
        public @bind NonTerminal(@bind.arg("Word") String name,
                                 @bind.arg("RHS") Seq[][] sequences) {
            this.name = name;
            this.sequences = sequences;
        }
        public Element build(MetaGrammar.Context cx) { return cx.get(name); }
    }

    public static class AnonUn extends Un {
        public @bind.as("(") AnonUn(Seq[][] sequences) {
            this.sequences = sequences;
        }
        public Element build(MetaGrammar.Context cx) {
            Union ret = new Union();
            build(cx, ret);
            return ret;
        }
    }

    //public static @bind.as void range(char c) { }
    public static class Range {
        public @bind.as("range") Range(char only) { first = only; last = only; }
        public @bind.as("-")     Range(char first, char last) { this.first = first; this.last = last; }
        public char first;
        public char last;
    }
    public static abstract class El {
        public String getLabel() { return null; }
        public String getOwnerTag() { return null; }
        public boolean drop() { return false; }
        public abstract Element build(MetaGrammar.Context cx);
    }
    public static class Drop extends El {
        public El e;
        public Drop(El e) { this.e = e; }
        public String getLabel() { return null; }
        public boolean drop() { return true; }
        public String getOwnerTag() { return e.getOwnerTag(); }
        public Element build(MetaGrammar.Context cx) { return e.build(cx); }
    }
    public static class Label extends El {
        public String label;
        public El e;
        public Label(String label, El e) { this.e = e; this.label = label; }
        public String getLabel() { return label; }
        public String getOwnerTag() { return e.getOwnerTag(); }
        public Element build(MetaGrammar.Context cx) { return e.build(cx); }
    }
    public static /*abstract*/ class Seq {
        HashSet<Seq> and = new HashSet<Seq>();
        HashSet<Seq> not = new HashSet<Seq>();
        El[] elements;
        El follow;
        String tag = null;
        boolean lame;
        public Seq(El e) { this(new El[] { e }); }
        public Seq(El[] elements) { this.elements = elements; }
        public Seq tag(String tag) { this.tag = tag; return this; }
        public Seq follow(El follow) { this.follow = follow; return this; }
        public Seq dup() {
            Seq ret = new Seq(elements);
            ret.and.addAll(and);
            ret.not.addAll(not);
            ret.follow = follow;
            ret.tag = tag;
            return ret;
        }
        public Seq and(Seq s) { and.add(s); s.lame = true; return this; }
        public Seq andnot(Seq s) { not.add(s); s.lame = true; return this; }
        public Seq separate(El sep) {
            El[] elements = new El[this.elements.length * 2 - 1];
            for(int i=0; i<this.elements.length; i++) {
                elements[i*2]   = this.elements[i];
                if (i<this.elements.length-1)
                    elements[i*2+1] = new Drop(sep);
            }
            this.elements = elements;
            return this;
        }
        public Sequence build(MetaGrammar.Context cx, Union u, boolean lame) {
            Sequence ret = build0(cx, lame || this.lame);
            for(Seq s : and) { Sequence dork = s.build(cx, u, true); ret = ret.and(dork); }
            for(Seq s : not) { Sequence dork = s.build(cx, u, true); ret = ret.not(dork); }
            u.add(ret);
            ret.lame = lame;
            return ret;
        }
        public Sequence build0(MetaGrammar.Context cx, boolean lame) {
            boolean unwrap = false;
            boolean dropAll = lame;
            if (tag!=null && tag.equals("[]")) unwrap  = true;
            if (tag!=null && "()".equals(tag)) dropAll = true;
            Object[] labels = new Object[elements.length];
            boolean[] drops = new boolean[elements.length];
            Element[] els = new Element[elements.length];
            for(int i=0; i<elements.length; i++) {
                labels[i] = elements[i].getLabel();
                drops[i]  = elements[i].drop();
                els[i] = elements[i].build(cx);
                if (elements[i].getOwnerTag() != null)
                    tag = elements[i].getOwnerTag();
            }
            Sequence ret = null;
            if (dropAll)     ret = Sequence.drop(els, false);
            else if (unwrap) ret = Sequence.unwrap(els, cx.rm.repeatTag(), drops);
            else if (tag!=null) {
                ret = cx.rm.resolveTag(tag, cx.cnt, els, labels, drops);
            } else {
                ret = cx.rm.tryResolveTag(tag, cx.cnt, els, labels, drops);
                if (ret == null) {
                    int idx = -1;
                    for(int i=0; i<els.length; i++)
                        if (!drops[i])
                            if (idx==-1) idx = i;
                            else throw new Error("multiple non-dropped elements in sequence: " + Sequence.drop(els,false));
                    if (idx != -1) ret = Sequence.singleton(els, idx);
                    else           ret = Sequence.drop(els, false);
                }
            }
            if (this.follow != null)
                ret.follow = infer(this.follow.build(cx));
            ret.lame = this.lame;
            return ret;
        }
    }
    public static @bind.as("&")   Seq  and(Seq s,         El[] elements) { return s.and(seq(elements)); }
    public static @bind.as("&~")  Seq  andnot(Seq s,      El[] elements) { return s.andnot(seq(elements)); }
    public static @bind.as("->")  Seq  arrow(Seq s, El e)                { return s.follow(e); }
    public static @bind.as("::")  Seq  tag(String tagname, Seq s)        { return s.tag(tagname); }
    public static @bind.as("/")   Seq  slash(Seq s, El e)                { return s.separate(e); }

    public static @bind.as("ps")  Seq  seq(El[] elements)                { return new Seq(elements); }
    public static @bind.as        Seq  psx(Seq s)                        { return s; }
    public static @bind.as(":")   El   colon(String s, El e)             { return new Label(s, e); }
    public static @bind.as(")")   void close(String foo)                 { throw new Error("not supported"); }
    public static @bind.as("()")  El   epsilon()                         { return new Constant(Union.epsilon); }

    public static @bind.as("nonTerminal") class NonTerminalReference extends El {
        public @bind.arg String nonTerminal;
        public Element build(MetaGrammar.Context cx) {
            return cx.get(nonTerminal);
        }
    }

    public static class StringLiteral        extends Constant {
        public @bind.as("literal") StringLiteral(String string) { super(CharRange.string(string)); }
        public boolean drop() { return true; }
    }

    public static                     class CharClass            extends El {
        Range[] ranges;
        public @bind.as("[") CharClass(Range[] ranges) { this.ranges = ranges; }
        public Element build(MetaGrammar.Context cx) {
            edu.berkeley.sbp.util.Range.Set set = new edu.berkeley.sbp.util.Range.Set();
            for(Range r : ranges)
                set.add(r.first, r.last);
            return CharRange.set(set);
        }
    }

    public static @bind.as("{")           class XTree                 extends El {
        public @bind.arg Seq body;
        public Element build(MetaGrammar.Context cx) {
            throw new Error();
        }
    }

    public static class Rep extends El {
        public El e, sep;
        public boolean zero, many, max;
        public Rep(El e, El sep, boolean zero, boolean many, boolean max) {
            this.e = e; this.sep = sep; this.zero = zero; this.many = many; this.max = max;}
        public Element build(MetaGrammar.Context cx) {
            return (!max)
                ? Sequence.repeat(e.build(cx),        zero, many, sep==null ? null : sep.build(cx), cx.rm.repeatTag())
                : sep==null
                ? Sequence.repeatMaximal(infer(e.build(cx)), zero, many,                                   cx.rm.repeatTag())
                : Sequence.repeatMaximal(e.build(cx),                    zero, many, infer(sep.build(cx)), cx.rm.repeatTag());
        }
    }
    public static class Constant extends El {
        Element constant;
        public Constant(Element constant) { this.constant = constant; }
        public Element build(MetaGrammar.Context cx) { return constant; }
    }
    public abstract static class PostProcess extends El {
        El e;
        public PostProcess(El e) { this.e = e; }
        public Element build(MetaGrammar.Context cx) { return postProcess(e.build(cx)); }
        public abstract Element postProcess(Element e);
    }

    // FIXME: it would be nice if we could hoist this into "Rep"
    public static @bind.as("++")  El plusmax(final El e)                     { return new Rep(e, null, false, true, true); }
    public static @bind.as("+")   El plus(final El e)                        { return new Rep(e, null, false, true, false); }
    public static @bind.as("++/") El plusmaxfollow(final El e, final El sep) { return new Rep(e, sep,  false, true, true); }
    public static @bind.as("+/")  El plusfollow(final El e, final El sep)    { return new Rep(e, sep,  false, true, false); }
    public static @bind.as("**")  El starmax(final El e)                     { return new Rep(e, null, true,  true, true); }
    public static @bind.as("*")   El star(final El e)                        { return new Rep(e, null, true,  true, false); }
    public static @bind.as("**/") El starmaxfollow(final El e, final El sep) { return new Rep(e, sep,  true,  true, true); }
    public static @bind.as("*/")  El starfollow(final El e, final El sep)    { return new Rep(e, sep,  true,  true, false); }
    public static @bind.as("?")   El question(final El e)                    { return new Rep(e, null, true,  true, false); }

    public static @bind.as("!")   El bang(final El e)                        { return new Drop(e); }

    public static @bind.as("^")   El caret(final String s) {
        return new Drop(new Constant(CharRange.string(s)) {
                public String getOwnerTag() { return s; }
            });
    }

    public static @bind.as("~")   El tilde(final El e) {
        return new PostProcess(e) {
                public Element postProcess(Element e) {
                    return infer((Topology<Character>)Atom.toAtom(e).complement()); 
                } }; }

    public static @bind.as("^^")  void doublecaret(final El e)                 { throw new Error("not implemented"); }

    //public static @bind.as("(")   El subexpression(Seq[][] rhs)                { return new NonTerminal(rhs); }

    public static @bind.as("Word")    String word(String s) { return s; }
    public static @bind.as("Quoted")  String quoted(String s) { return s; }
    public static @bind.as("escaped") String c(char c) { return c+""; }
    public static @bind.as("\"\"")    String emptystring() { return ""; }
    public static @bind.as("\n")      String retur() { return "\n"; }
    public static @bind.as("\r")      String lf() { return "\r"; }

    static Atom infer(Element e)  { return infer((Topology<Character>)Atom.toAtom(e)); }
    static Atom infer(Topology<Character> t) { return new CharRange(new CharTopology(t)); }
}
